package jez04.structure.test;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.nio.file.FileVisitResult;
import java.nio.file.FileVisitor;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.List;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import cz.vsb.fei.lab.App;
import cz.vsb.fei.lab.Score;
import cz.vsb.fei.lab.ScoreNotLoaded;

class ClassStructureTest {

	static StructureHelper helper = new StructureHelper();
	private static String className = "Score";

	@BeforeAll
	static void printClassNames() {
		for (String className : helper.getNameOfAllClasses()) {
			System.out.println(className);
		}
	}

	@Test
	void scoreExistenceTest() {
		helper.classExist(className);
		Class<?> c = helper.getClass(className);
		helper.hasProperty(c, "name", String.class, false);
		helper.hasProperty(c, "points", int.class, false);
		helper.hasMethod(c, "generate", Score.class);
	}

	@Test
	void scoreNotLoadedExistenceTest() {
		helper.classExist(className);
		Class<?> c = helper.getClass("ScoreNotLoaded");
		helper.hasProperty(c, "count", int.class, false);
		helper.hasExtends(c, Exception.class);
	}

	@Test
	void generateScoresTest() throws Exception {
		App app = new App();
		assertTrue(app.generateScores(10).size() == 10);
	}

	@Test
	void loadScoresFromStreamTest() throws Exception {
		App app = new App();
		assertTrue(app.loadScoresFromStream(new InputStreamReader(App.class.getResourceAsStream("/bestScore-ok.csv")))
				.size() > 1);
	}

	@Test
	void loadScoresFromStreamExceptionTest() throws Exception, IllegalAccessException, InvocationTargetException,
			NoSuchMethodException, SecurityException {
		App app = new App();
		try {
			app.loadScoresFromStream(new InputStreamReader(App.class.getResourceAsStream("/bestScores-err.csv")))
					.size();
			fail("Exception not throwen!");
		} catch (Exception e) {
			assertEquals(ScoreNotLoaded.class, e.getClass(), "Excpectin exception of type ScoreNotLoaded");
			int count = (int) ScoreNotLoaded.class.getDeclaredMethod("getCount").invoke(e);
			assertTrue(count > 0, "Info about already parsed line missing");
		}
	}

	@Test
	void loadScoresTest() throws Exception {
		App app = new App();
		List<Score> scores = app.loadScores();
		assertNotNull(scores);
		assertTrue(scores.size() > 0, "Collection is empty");
	}

	@Test
	void loadScoresDisableCsvFailTest() throws Exception {
		try {
			Files.walkFileTree(Paths.get("."), new FileVisitor<Path>() {

				@Override
				public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
					if (file.getFileName().toString().toLowerCase().endsWith(".csv")) {
						file.toFile().renameTo(Paths.get(file.toString() + "-DISABLE").toFile());
						return FileVisitResult.CONTINUE;
					}
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
					return FileVisitResult.CONTINUE;
				}
			});

			App app = new App();
			List<Score> scores = app.loadScores();
			assertNotNull(scores);
			assertTrue(scores.isEmpty(), "Collection have to be empty.");
		} finally {
			Files.walkFileTree(Paths.get("."), new FileVisitor<Path>() {

				@Override
				public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
					if (file.getFileName().toString().endsWith("-DISABLE")) {
						String fileName = file.toString();
						file.toFile().renameTo(
								Paths.get(fileName.substring(0, fileName.length() - "-DISABLE".length())).toFile());
						return FileVisitResult.CONTINUE;
					}
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
					return FileVisitResult.CONTINUE;
				}
			});

		}
	}

	@Test
	void saveScoresTest() throws IOException, NoSuchMethodException, SecurityException, InstantiationException,
			IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		App app = new App();
		app.saveScores(List.of(createScore("test", 1), createScore("test2", 15)), "test.csv");
		assertEquals(2, Files.lines(Paths.get("test.csv")).count(),
				"Expecting existent of file test.csv with 2 lines.");
	}

	private Score createScore(String name, int points) throws NoSuchMethodException, SecurityException,
			InstantiationException, IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		Constructor constructor = Score.class.getConstructor(String.class, int.class);
		return (Score) constructor.newInstance(name, points);
	}

	@Test
	void saveScoresDirsTest() throws IOException, NoSuchMethodException, SecurityException, InstantiationException,
			IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		App app = new App();
		app.saveScores(List.of(createScore("test", 1), createScore("test2", 15)), "test/subtest/test.csv");
		assertEquals(2, Files.lines(Paths.get("test", "subtest", "test.csv")).count(),
				"Expecting existent of file test.csv with 2 lines.");
	}

}
